from arsenic.session import Session, CompatSession


class Browser:
    defaults = {}
    session_class = Session

    def __init__(self, **overrides):
        self.capabilities = {**self.defaults, **overrides}


class Firefox(Browser):
    defaults = {
        'browserName': 'firefox',
        'marionette': True,
        'acceptInsecureCerts': True,
    }


class Chrome(Browser):
    session_class = CompatSession
    defaults = {
        'browserName': 'chrome',
    }


class PhantomJS(Browser):
    session_class = CompatSession
    defaults = {
        'browserName': 'phantomjs',
        'version': '',
        'platform': 'ANY',
        'javascriptEnabled': True,
    }


class InternetExplorer(Browser):
    session_class = CompatSession
    defaults = {
        'browserName': 'internet explorer',
        'version': '',
        'platform': 'WINDOWS',
    }

IE = InternetExplorer



import abc
from enum import Enum
from typing import Sequence, Any, Iterator, List, Dict, Optional

import attr

from arsenic import constants
from arsenic.session import Element


@attr.s
class Action:
    source = attr.ib()
    payload = attr.ib()


class Tick:
    def __init__(self, *actions: Action):
        self.actions: Dict['Device', 'Action'] = {
            action.source: action
            for action in actions
        }

    def __and__(self, other: 'Tick') -> 'Tick':
        overlap = set(self.actions.keys()) & set(other.actions.keys())
        if overlap:
            raise ValueError(
                f'Devices {overlap} have more than one action in this tick'
            )
        return Tick(*self.actions.values(), *other.actions.values())

    def encode(self, device: 'Device') -> Dict[str, Any]:
        if device not in self.actions:
            return {
                'type': 'pause',
                'duration': 0,
            }
        else:
            return self.actions[device].payload


class DeviceType(Enum):
    keyboard = 'key'
    pointer = 'pointer'


class PointerType(Enum):
    mouse = 'mouse'
    pen = 'pen'
    touch = 'touch'


class Button(Enum):
    left = 0
    middle = 1
    right = 2


class Device(metaclass=abc.ABCMeta):
    type: DeviceType = abc.abstractproperty()

    def __init__(self, device_id: Optional[str]=None):
        self.device_id = device_id

    def info(self, index: int) -> Dict[str, Any]:
        device_id = self.device_id or f'{self.type.value}{index}'
        return {
            'id': device_id,
            'type': self.type.value,
        }

    def _tick(self, **payload: Any) -> Tick:
        return Tick(Action(self, payload))

    def pause(self, duration: int) -> Tick:
        return self._tick(
            type='pause',
            duration=duration,
        )


class Pointer(Device, metaclass=abc.ABCMeta):
    type = DeviceType.pointer
    pointer_type: PointerType = abc.abstractproperty()

    def info(self, index: int) -> Dict[str, Any]:
        return {
            'parameters': {'pointerType': self.pointer_type.value},
            **super().info(index)
        }

    def move_to(self, element: Element, duration: int=250) -> Tick:
        return self._tick(
            type='pointerMove',
            duration=duration,
            origin={constants.WEB_ELEMENT: element.id},
            x=0,
            y=0,
        )

    def move_by(self, x: int, y: int, duration: int=250) -> Tick:
        return self._tick(
            type='pointerMove',
            duration=duration,
            origin='pointer',
            x=x,
            y=y,
        )

    def down(self) -> Tick:
        return self._tick(
            type='pointerDown',
            duration=0,
            button=0,
        )

    def up(self) -> Tick:
        return self._tick(
            type='pointerUp',
            duration=0,
            button=0,
        )


class Mouse(Pointer):
    pointer_type = PointerType.mouse

    def down(self, button: Button=Button.left) -> Tick:
        return self._tick(
            type='pointerDown',
            duration=0,
            button=button.value
        )

    def up(self, button: Button=Button.left) -> Tick:
        return self._tick(
            type='pointerUp',
            duration=0,
            button=button.value
        )


class Pen(Pointer):
    pointer_type = PointerType.pen


class Touch(Pointer):
    pointer_type = PointerType.touch


class Keyboard(Device):
    type = DeviceType.keyboard

    def down(self, key: str) -> Tick:
        return self._tick(
            type='keyDown',
            value=key
        )

    def up(self, key: str) -> Tick:
        return self._tick(
            type='keyUp',
            value=key
        )


def gather_devices(ticks: Sequence[Tick]) -> Iterator[Device]:
    found = set()
    for tick in ticks:
        devices = set(tick.actions.keys())
        for device in devices - found:
            yield device
        found.update(devices)


def chain(*ticks: Tick) -> Dict[str, List[Dict[str, Any]]]:
    devices = list(gather_devices(ticks))
    return {
        'actions': [
            {
                **device.info(index),
                'actions': [tick.encode(device) for tick in ticks],
            } for index, device in enumerate(devices, start=1)
        ]
    }



import socket

import attr


def free_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind(('0.0.0.0', 0))
        sock.listen(5)
        return sock.getsockname()[1]


def px_to_int(value: str) -> int:
    original = value
    if value.endswith('px'):
        value = value[:-2]
    if value.isdigit():
        return int(value)
    else:
        raise ValueError(f'{original!r} is not an int or <int>px value')


@attr.s
class Rect:
    x: float = attr.ib()
    y: float = attr.ib()
    width: float = attr.ib()
    height: float = attr.ib()



import abc
import base64
from typing import Dict

import attr


class Auth(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_headers(self) -> Dict[str, str]:
        raise NotImplementedError()


@attr.s
class BasicAuth(Auth):
    username = attr.ib()
    password = attr.ib()

    def get_headers(self):
        raw_token = f'{self.username}:{self.password}'
        token = base64.b64encode(raw_token.encode('ascii')).decode('ascii')
        return {
            'Authorization': f'Basic {token}'
        }



from functools import partial
from io import BytesIO
from pathlib import Path
from typing import Awaitable, Callable, Any, List, Dict, Tuple, Iterator

import attr
import base64

from arsenic import errors, constants
from arsenic.connection import Connection
from arsenic.errors import NoSuchElement, OperationNotSupported
from arsenic.utils import Rect, px_to_int

UNSET = object()


def escape_value(value: str) -> str:
    if '"' in value and "'" in value:
        parts = value.split('"')
        result = ['concat(']
        for part in parts:
            result.append(f'"{part}"')
            result.append(', \'"\', ')
        result = result[0:-1]
        if value.endswith('"'):
            return ''.join(result) + ')'
        else:
            return ''.join(result[:-1]) + ')'
    elif '"' in value:
        return f"'{value}'"
    else:
        return f'"{value}"'


class RequestHelpers:
    async def _request(self, *, url: str, method: str, data=None, raw=False):
        status, data = await self.connection.request(url=url, method=method, data=data)
        self._check_response_error(status, data)
        if raw:
            return data
        if data:
            return self._unwrap(data.get('value', None))

    def _check_response_error(self, status: int, data: Any) -> None:
        if status >= 400:
            errors.raise_exception(data, status)

    def _unwrap(self, value):
        """
        Unwrap a value returned from a webdriver. Specifically this means trying to
        extract the element ID of a web element or a list of web elements.
        """
        if isinstance(value, dict) and ('ELEMENT' in value or constants.WEB_ELEMENT in value):
            wrapped_id = value.get('ELEMENT', None)
            if wrapped_id:
                return value['ELEMENT']
            else:
                return value[constants.WEB_ELEMENT]

        elif isinstance(value, list):
            return list(self._unwrap(item) for item in value)
        else:
            return value


class Element(RequestHelpers):
    def __init__(self, id: str, connection: Connection, session: 'Session'):
        self.id = id
        self.connection = connection
        self.session = session

    async def get_text(self) -> str:
        return await self._request(
            url='/text',
            method='GET'
        )

    async def send_keys(self, keys: str):
        await self._request(
            url='/value',
            method='POST',
            data={
                'value': list(keys),
                'text': keys,
            }
        )

    async def send_file(self, path: Path):
        path = await self.session.connection.upload_file(path)
        await self.send_keys(str(path))

    async def clear(self):
        await self._request(
            url='/clear',
            method='POST'
        )

    async def click(self):
        await self._request(
            url='/click',
            method='POST'
        )

    async def is_displayed(self) -> bool:
        return await self._request(
            url='/displayed',
            method='GET'
        )

    async def is_enabled(self) -> bool:
        return await self._request(
            url='/enabled',
            method='GET'
        )

    async def get_attribute(self, name: str) -> str:
        return await self._request(
            url=f'/attribute/{name}',
            method='GET'
        )

    async def get_css_value(self, name: str) -> str:
        return await self._request(
            url=f'/css/{name}',
            method='GET',
        )

    async def select_by_value(self, value: str):
        value = escape_value(value)
        option = await self.get_element(f'option[value={value}]')
        await option.click()

    async def get_element(self, selector: str) -> 'Element':
        element_id = await self._request(
            url='/element',
            method='POST',
            data={
                'using': 'css selector',
                'value': selector,
            }
        )
        return self.session.create_element(element_id)

    async def get_elements(self, selector: str) -> List['Element']:
        element_ids = await self._request(
            url='/elements',
            method='POST',
            data={
                'using': 'css selector',
                'value': selector,
            }
        )
        return [self.session.create_element(element_id) for element_id in element_ids]

    async def get_rect(self):
        data = await self._request(
            url='/rect',
            method='GET',
        )
        return Rect(data['x'], data['y'], data['width'], data['height'])


TCallback = Callable[..., Awaitable[Any]]
TWaiter = Callable[[int, TCallback], Awaitable[Any]]


class Session(RequestHelpers):
    element_class = Element

    def __init__(self, connection: Connection, wait: TWaiter, driver, bind: str=''):
        self.connection = connection
        self.bind = bind
        self.wait = wait
        self.driver = driver

    async def get(self, url: str):
        await self._request(
            url='/url',
            method='POST',
            data={
                'url': self.bind + url
            }
        )

    async def get_url(self):
        return await self._request(
            url='/url',
            method='GET'
        )

    async def get_page_source(self) -> str:
        return await self._request(
            url='/source',
            method='GET'
        )

    async def get_element(self, selector: str) -> Element:
        element_id = await self._request(
            url='/element',
            method='POST',
            data={
                'using': 'css selector',
                'value': selector
            }
        )
        return self.create_element(element_id)

    async def get_elements(self, selector: str) -> List[Element]:
        result = await self._request(
            url='/elements',
            method='POST',
            data={
                'using': 'css selector',
                'value': selector
            }
        )
        return [self.create_element(element_id) for element_id in result]

    async def wait_for_element(self, timeout: int, selector: str) -> Element:
        return await self.wait(
            timeout,
            partial(self.get_element, selector),
            NoSuchElement
        )

    async def wait_for_element_gone(self, timeout: int, selector: str):
        async def callback():
            try:
                await self.get_element(selector)
            except NoSuchElement:
                return True
            else:
                return False
        return await self.wait(timeout, callback)

    async def add_cookie(self, name: str, value: str, *, path: str=UNSET, domain: str=UNSET, secure: bool=UNSET, expiry: int=UNSET):
        cookie = {
            'name': name,
            'value': value
        }
        if path is not UNSET:
            cookie['path'] = path
        if domain is not UNSET:
            cookie['domain'] = domain
        if secure is not UNSET:
            cookie['secure'] = secure
        if expiry is not UNSET:
            cookie['expiry'] = expiry
        await self._request(
            url='/cookie',
            method='POST',
            data={
                'cookie': cookie
            }
        )

    async def get_cookie(self, name: str):
        return await self._request(
            url=f'/cookie/{name}',
            method='GET'
        )

    async def get_all_cookies(self):
        return await self._request(
            url='/cookie',
            method='GET'
        )

    async def delete_cookie(self, name: str):
        await self._request(
            url=f'/cookie/{name}',
            method='DELETE'
        )

    async def delete_all_cookies(self):
        await self._request(
            url='/cookie',
            method='DELETE'
        )

    async def execute_script(self, script: str, *args: Any):
        return await self._request(
            url='/execute/sync',
            method='POST',
            data={
                'script': script,
                'args': list(args)
            }
        )

    async def set_window_size(self, width: int, height: int, handle: str='current'):
        return await self._request(
            url='/window/rect',
            method='POST',
            data={
                'width': width,
                'height': height,
                'windowHandle': handle,
            }
        )

    async def get_window_size(self, handle: str='current') -> Tuple[int, int]:
        return await self._request(
            url='/window/rect',
            method='GET',
            data={
                'windowHandle': handle,
            }
        )

    async def get_alert_text(self) -> str:
        return await self._request(
            url='/alert/text',
            method='GET'
        )

    async def send_alert_text(self, value: str):
        return await self._request(
            url='/alert/text',
            method='POST',
            data={
                'text': value
            }
        )

    async def dismiss_alert(self):
        return await self._request(
            url='/alert/dismiss',
            method='POST'
        )

    async def accept_alert(self):
        return await self._request(
            url='/alert/accept',
            method='POST'
        )

    async def perform_actions(self, actions: Dict[str, Any]):
        return await self._request(
            url='/actions',
            method='POST',
            data=actions
        )

    async def get_screenshot(self) -> BytesIO:
        return BytesIO(base64.b64decode(await self._request(
            url='/screenshot',
            method='GET'
        )))

    async def close(self):
        await self._request(
            url='',
            method='DELETE'
        )

    def create_element(self, element_id):
        return self.element_class(
            element_id,
            self.connection.prefixed(f'/element/{element_id}'),
            self
        )


class CompatRequestHelpers(RequestHelpers):
    def _check_response_error(self, status: int, data: Any):
        if 'status' in data and data['status'] != 0:
            errors.raise_exception(data, status)


class CompatElement(CompatRequestHelpers, Element):
    async def get_rect(self):
        location = await self._request(
            url='/location',
            method='GET'
        )
        width = await self.get_css_value('width')
        height = await self.get_css_value('height')
        return Rect(location['x'], location['y'], px_to_int(width), px_to_int(height))


class CompatSession(CompatRequestHelpers, Session):
    element_class = CompatElement

    async def set_window_size(self, width: int, height: int,
                              handle: str = 'current'):
        return await self._request(
            url=f'/window/{handle}/size',
            method='POST',
            data={
                'width': width,
                'height': height
            }
        )

    async def get_window_size(self, handle: str = 'current'):
        return await self._request(
            url=f'/window/{handle}/size',
            method='GET'
        )

    async def execute_script(self, script, *args):
        return await self._request(
            url='/execute',
            method='POST',
            data={
                'script': script,
                'args': list(args)
            }
        )

    async def perform_actions(self, actions: Dict[str, Any]):
        for url, method, data in transform_legacy_actions(actions['actions']):
            await self._request(
                url=url,
                method=method,
                data=data,
            )

    async def get_screenshot(self) -> BytesIO:
        return BytesIO(base64.b64decode(await self._request(
            url='/screenshot',
            method='GET'
        )))


def _pointer_down(device, action):
    del action['duration']
    url = '/buttondown' if device['parameters']['pointerType'] == 'mouse' else '/touch/down'
    return url, 'POST', action


def _pointer_up(device, action):
    del action['duration']
    url = '/buttonup' if device['parameters']['pointerType'] == 'mouse' else '/touch/up'
    return url, 'POST', action


def _pointer_move(device, action):
    del action['duration']
    url = '/moveto' if device['parameters']['pointerType'] == 'mouse' else '/touch/move'
    origin = action['origin']
    if origin == 'pointer':
        data = {
            'xoffset': action['x'],
            'yoffset': action['y'],
        }
    elif constants.WEB_ELEMENT in origin:
        data = {
            'element': origin[constants.WEB_ELEMENT],
        }
    else:
        raise OperationNotSupported(f'Cannot move using origin {origin}')
    return url, 'POST', data


def _pause(device, action):
    return None


def key_down(device, action):
    return '/keydown', 'POST', {''}


legacy_actions = {
    ('pointer', 'pointerDown'): _pointer_down,
    ('pointer', 'pointerUp'): _pointer_up,
    ('pointer', 'pointerMove'): _pointer_move,
    ('pointer', 'pause'): _pause,
    ('key', 'pause'): _pause,
}


@attr.s
class LegacyAction:
    device = attr.ib()
    action = attr.ib()


def get_legacy_actions(devices: List[Dict[str, Any]]) -> Iterator[LegacyAction]:
    i = 0
    while devices:
        for device in devices:
            action = device['actions'].pop(0)
            i += 1
            yield LegacyAction(device, action)
        devices = [
            device for device in devices if device['actions']
        ]


def transform_legacy_actions(devices: List[Dict[str, Any]]) -> Iterator[Tuple[str, str, Dict[str, Any]]]:
    for legacy_action in get_legacy_actions(devices):
        device_type = legacy_action.device['type']
        action_type = legacy_action.action['type']
        device = {key: value for key, value in legacy_action.device.items() if key != 'type'}
        action = {key: value for key, value in legacy_action.action.items() if key != 'type'}
        try:
            handler = legacy_actions[(device_type, action_type)]
        except KeyError:
            raise OperationNotSupported(
                f'Unsupported action {action_type} for device_type {device_type}'
            )
        action = handler(device, action)
        if action is not None:
            yield action



from setuptools import setup, find_packages


setup(
    version='1.0.0.dev6',
    name='arsenic',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    install_requires=[
        'attrs>=17.1.0',
        'structlog',
        'aiohttp>=2',
    ],
    license='APLv2',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6'
    ]
)



from typing import Union, Dict, Type, Any

from structlog import get_logger

log = get_logger()


class ArsenicError(Exception):
    pass


class OperationNotSupported(ArsenicError):
    pass


class WebdriverError(ArsenicError):
    def __init__(self, message, screen, stacktrace):
        self.message = message
        self.screen = screen
        self.stacktrace = stacktrace
        super().__init__(message)


class UnknownArsenicError(ArsenicError):
    pass


class ArsenicTimeout(ArsenicError):
    pass


CODES: Dict[Union[str, int], Type[WebdriverError]] = {}


def get(error_code: Union[str, int]) -> Type[WebdriverError]:
    return CODES.get(error_code, UnknownArsenicError)


def create(error_name: str, *error_codes: int) -> Type[WebdriverError]:
    name = ''.join(bit.capitalize() for bit in error_name.split(' '))
    cls: Type[WebdriverError] = type(name, (WebdriverError,), {})
    CODES[error_name] = cls
    for code in error_codes:
        CODES[code] = cls
    return cls


NoSuchElement = create('no such element', 7)
NoSuchFrame = create('no such frame', 8)
UnknownCommand = create('unknown command', 9)
StaleElementReference = create('stale element reference', 10)
ElementNotVisible = create('element not visible', 11)
InvalidElementState = create('invalid element state', 12)
UnknownError = create('unknown error', 13)
ElementNotInteractable = create('element not interactable')
ElementIsNotSelectable = create('element is not selectable', 15)
JavascriptError = create('javascript error', 17)
Timeout = create('timeout', 21)
NoSuchWindow = create('no such window', 23)
InvalidCookieDomain = create('invalid cookie domain', 24)
UnableToSetCookie = create('unable to set cookie', 25)
UnexpectedAlertOpen = create('unexpected alert open', 26)
NoSuchAlert = create('no such alert', 27)
ScriptTimeout = create('script timeout', 28)
InvalidElementCoordinates = create('invalid element coordinates', 29)
IMENotAvailable = create('ime not available', 30)
IMEEngineActivationFailed = create('ime engine activation failed', 31)
InvalidSelector = create('invalid selector', 32)
MoveTargetOutOfBounds = create('move target out of bounds', 34)


def raise_exception(data: Dict[str, Any], status: int):
    error = None
    if 'status' in data:
        error = data['status']
    elif 'error' in data:
        error = data['error']
    elif 'state' in data:
        error = data['state']

    if 'value' in data and isinstance(data['value'], dict):
        data = data['value']
    if error is None and 'error' in data:
        error = data['error']
    message = data.get('message', None)
    stacktrace = data.get('stacktrace', None)
    screen = data.get('screen', None)
    exception_class = get(error)
    log.error('error', type=exception_class, message=message, stacktrace=stacktrace, data=data, status=status)
    raise exception_class(message, screen, stacktrace)



"""
This module contains special keys as defined in the web driver specification.

They can be used in the send_keys API using string formatting.
"""
NULL = '\ue000'
CANCEL = '\ue001'
HELP = '\ue002'
BACKSPACE = '\ue003'
TAB = '\ue004'
CLEAR = '\ue005'
RETURN = '\ue006'
ENTER = '\ue007'
SHIFT = '\ue008'
CONTROL = '\ue009'
ALT = '\ue00a'
PAUSE = '\ue00b'
ESCAPE = '\ue00c'
SPACE = '\ue00d'
PAGE_UP = '\ue00e'
PAGE_DOWN = '\ue00f'
END = '\ue010'
HOME = '\ue011'
LEFT = '\ue012'
UP = '\ue013'
RIGHT = '\ue014'
DOWN = '\ue015'
INSERT = '\ue016'
DELETE = '\ue017'
SEMICOLON = '\ue018'
EQUALS = '\ue019'

NUMPAD0 = '\ue01a'
NUMPAD1 = '\ue01b'
NUMPAD2 = '\ue01c'
NUMPAD3 = '\ue01d'
NUMPAD4 = '\ue01e'
NUMPAD5 = '\ue01f'
NUMPAD6 = '\ue020'
NUMPAD7 = '\ue021'
NUMPAD8 = '\ue022'
NUMPAD9 = '\ue023'
MULTIPLY = '\ue024'
ADD = '\ue025'
SEPARATOR = '\ue026'
SUBTRACT = '\ue027'
DECIMAL = '\ue028'
DIVIDE = '\ue029'

F1 = '\ue031'
F2 = '\ue032'
F3 = '\ue033'
F4 = '\ue034'
F5 = '\ue035'
F6 = '\ue036'
F7 = '\ue037'
F8 = '\ue038'
F9 = '\ue039'
F10 = '\ue03a'
F11 = '\ue03b'
F12 = '\ue03c'

META = '\ue03d'
COMMAND = '\ue03d'



import abc
import asyncio
import os
import re
from asyncio.subprocess import DEVNULL, PIPE
from distutils.version import StrictVersion
from functools import partial
from typing import List, TextIO, Optional

import attr
import sys
from aiohttp import ClientSession

from arsenic.connection import Connection, RemoteConnection
from arsenic.utils import free_port
from arsenic.webdriver import WebDriver
from arsenic.http import Auth, BasicAuth


async def stop_process(process):
    process.terminate()
    try:
        await asyncio.wait_for(process.communicate(), 1)
    except asyncio.futures.TimeoutError:
        process.kill()
    try:
        await asyncio.wait_for(process.communicate(), 1)
    except asyncio.futures.TimeoutError:
        pass


def sync_factory(func):
    async def sync():
        func()
    return sync


async def tasked(coro):
    return await asyncio.get_event_loop().create_task(coro)


def check_event_loop():
    if sys.platform == 'win32' and isinstance(asyncio.get_event_loop(), asyncio.SelectorEventLoop):
        raise ValueError(
            'SelectorEventLoop is not supported on Windows, use asyncio.ProactorEventLoop instead.'
        )


async def run_subprocess(cmd: List[str]) -> str:
    check_event_loop()
    process = await asyncio.create_subprocess_exec(
        *cmd,
        stdout=PIPE,
        stderr=PIPE
    )
    out, err = await process.communicate()
    if process.returncode != 0:
        raise Exception(err)
    else:
        return out.decode('utf-8')


async def subprocess_based_service(cmd: List[str],
                                   service_url: str,
                                   log_file: TextIO) -> WebDriver:
    check_event_loop()
    closers = []
    try:
        if log_file is os.devnull:
            log_file = DEVNULL
        process = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=log_file,
            stderr=log_file,
        )
        closers.append(partial(stop_process, process))
        session = ClientSession()
        closers.append(sync_factory(session.close))
        count = 0
        while True:
            try:
                await tasked(session.request(
                    url=service_url + '/status',
                    method='GET'
                ))
                break
            except:
                # TODO: make this better
                count += 1
                if count > 30:
                    raise Exception('not starting?')
                await asyncio.sleep(0.5)
        return WebDriver(
            Connection(session, service_url),
            closers,
        )
    except:
        for closer in reversed(closers):
            await closer()
        raise


class Service(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def start(self) -> WebDriver:
        raise NotImplementedError()


@attr.s
class Geckodriver(Service):
    log_file = attr.ib(default=sys.stdout)
    binary = attr.ib(default='geckodriver')
    version_check = attr.ib(default=True)

    _version_re = re.compile(r'geckodriver (\d+\.\d+)')

    async def _check_version(self):
        if self.version_check:
            output = await run_subprocess([self.binary, '--version'])
            match = self._version_re.search(output)
            if not match:
                raise ValueError(
                    'Could not determine version of geckodriver. To '
                    'disable version checking, set `version_check` to '
                    '`False`.'
                )
            version_str = match.group(1)
            version = StrictVersion(version_str)
            if version < StrictVersion('0.16.1'):
                raise ValueError(
                    f'Geckodriver version {version_str} is too old. 0.16.1 or '
                    f'higher is required. To disable version checking, set '
                    f'`version_check` to `False`.'
                )

    async def start(self):
        port = free_port()
        await self._check_version()
        return await subprocess_based_service(
            [self.binary, '--port', str(port)],
            f'http://localhost:{port}',
            self.log_file
        )


@attr.s
class Chromedriver(Service):
    log_file = attr.ib(default=sys.stdout)
    binary = attr.ib(default='chromedriver')

    async def start(self):
        port = free_port()
        return await subprocess_based_service(
            [self.binary, f'--port={port}'],
            f'http://localhost:{port}',
            self.log_file
        )


def auth_or_string(value):
    if isinstance(value, Auth):
        return value
    elif isinstance(value, str) and value.count(':') == 1:
        username, password = value.split(':')
        return BasicAuth(username, password)
    else:
        raise TypeError()


@attr.s
class Remote(Service):
    url: str = attr.ib()
    auth: Optional[Auth] = attr.ib(default=None, convert=attr.converters.optional(auth_or_string))

    async def start(self):
        closers = []
        headers = {}
        if self.auth:
            headers.update(self.auth.get_headers())
        try:
            session = ClientSession(headers=headers)
            closers.append(sync_factory(session.close))
            return WebDriver(RemoteConnection(session, self.url), closers)
        except:
            for closer in reversed(closers):
                await closer()
            raise


@attr.s
class PhantomJS(Service):
    log_file = attr.ib(default=sys.stdout)
    binary = attr.ib(default='phantomjs')

    async def start(self):
        port = free_port()
        return await subprocess_based_service(
            [self.binary, f'--webdriver={port}'],
            f'http://localhost:{port}/wd/hub',
            self.log_file,
        )
